/*
# Smart Alarm System with Sleep Tracking

This project uses an ESP32, speaker, joystick button, and the Blynk IoT platform to create a connected alarm clock.
It plays a fun melody when the alarm triggers, tracks sleep duration, and sends a notification to Make.com with the time you slept.

## Hardware Requirements
- ESP32 development board  
- Piezo buzzer or speaker  
- Joystick module with button  
- Power supply for ESP32  
- Smartphone with Blynk app installed

## Pin Connections
- Joystick Button: GPIO 32  
- Speaker: GPIO 26  

## Features
- Set alarm time remotely using the Blynk app  
- Play "Never Gonna Give You Up" melody as the alarm tone  
- Joystick button used to stop the alarm  
- Automatically records sleep start and wake-up time  
- Calculates total hours slept and sends a notification via Make.com webhook  
- Time synced using NTP  
- Supports UTC offset for local timezone adjustments  

## Video
- https://youtu.be/qRPGj36vkL0?t=51

## Credits
- Theodore_Dai_Maman #211541594  
- Omer_Dan #322952466  
- Amit_Kaminsky #207487661  

## Setup Instructions
1. Install required libraries in Arduino IDE:  
   - Blynk  
   - WiFi  
   - HTTPClient  
   - TimeLib  

2. Update the following in the code:  
   - `ssid` and `password` with your WiFi credentials  
   - `BLYNK_AUTH_TOKEN` with your Blynk auth token  
   - `makeWebhookUrl` with your Make.com webhook URL  
   - (Optional) Adjust `UTC_OFFSET` for your local timezone  

3. Upload the code to your ESP32  

## Notes
- Alarm only plays when the current time matches the set alarm time  
- You can set or disable the alarm remotely through the Blynk app  
- Sleep tracking is based on the time the alarm was activated and stopped  
- Melody can be customized in the `melody[]` array  
*/

#define BLYNK_TEMPLATE_ID           "Enter Template ID"
#define BLYNK_TEMPLATE_NAME         "Enter Template Name"
#define BLYNK_AUTH_TOKEN            "Enter Auth Token"
#define BLYNK_PRINT Serial

#include <WiFi.h>
#include <BlynkSimpleEsp32.h>
#include <HTTPClient.h>
#include <TimeLib.h>

// Include pitches.h for note definitions
#include "pitches.h"

// WiFi credentials
const char* ssid = "Enter SSID";
const char* password = "Enter Password";

// Pin definitions
const int JOYSTICK_BUTTON_PIN = 32;  // Joystick button pin
const int SPEAKER_PIN = 26;          // Speaker pin

// Time settings
const int UTC_OFFSET = 3;  // UTC+3 for your timezone

// Structure to store time information
struct TimeInfo {
  int hour;
  int minute;
  int day;
  int month;
  int year;
};

// Alarm variables  
int alarmHour = 0;
int alarmMinute = 0;
bool alarmEnabled = false;
bool isAlarmPlaying = false;
TimeInfo sleepTimeInfo = {0, 0, 0, 0, 0};  // Store sleep time components
bool sleepTimeSet = false;  // Flag to track if sleep time is set

// Make.com webhook URL for notifications
const char* makeWebhookUrl = "Enter Webhook URL";

// Never Gonna Give You Up - Rick Astley
// Notes of the melody followed by the duration.
// a 4 means a quarter note, 8 an eighteenth , 16 sixteenth, so on
// !!negative numbers are used to represent dotted notes,
// so -4 means a dotted quarter note, that is, a quarter plus an eighteenth!!
int melody[] = {
  NOTE_FS5,8, NOTE_E5,-4,
  NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_FS5,-8, NOTE_FS5,-8, NOTE_E5,-4, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,

  NOTE_E5,-8, NOTE_E5,-8, NOTE_D5,-8, NOTE_CS5,16, NOTE_B4,-8, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16, //18
  NOTE_D5,4, NOTE_E5,8, NOTE_CS5,-8, NOTE_B4,16, NOTE_A4,8, NOTE_A4,8, NOTE_A4,8, 
  NOTE_E5,4, NOTE_D5,2, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_FS5,-8, NOTE_FS5,-8, NOTE_E5,-4, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_A5,4, NOTE_CS5,8, NOTE_D5,-8, NOTE_CS5,16, NOTE_B4,8, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,

  NOTE_D5,4, NOTE_E5,8, NOTE_CS5,-8, NOTE_B4,16, NOTE_A4,4, NOTE_A4,8,  //23
  NOTE_E5,4, NOTE_D5,2, REST,4,
  REST,8, NOTE_B4,8, NOTE_D5,8, NOTE_B4,8, NOTE_D5,8, NOTE_E5,4, REST,8,
  REST,8, NOTE_CS5,8, NOTE_B4,8, NOTE_A4,-4, REST,4,
  REST,8, NOTE_B4,8, NOTE_B4,8, NOTE_CS5,8, NOTE_D5,8, NOTE_B4,8, NOTE_A4,4,
  REST,8, NOTE_A5,8, NOTE_A5,8, NOTE_E5,8, NOTE_FS5,8, NOTE_E5,8, NOTE_D5,8,
  
  REST,8, NOTE_A4,8, NOTE_B4,8, NOTE_CS5,8, NOTE_D5,8, NOTE_B4,8, //29
  REST,8, NOTE_CS5,8, NOTE_B4,8, NOTE_A4,-4, REST,4,
  NOTE_B4,8, NOTE_B4,8, NOTE_CS5,8, NOTE_D5,8, NOTE_B4,8, NOTE_A4,4, REST,8,
  REST,8, NOTE_E5,8, NOTE_E5,8, NOTE_FS5,4, NOTE_E5,-4, 
  NOTE_D5,2, NOTE_D5,8, NOTE_E5,8, NOTE_FS5,8, NOTE_E5,4, 
  NOTE_E5,8, NOTE_E5,8, NOTE_FS5,8, NOTE_E5,8, NOTE_A4,8, NOTE_A4,4,

  REST,-4, NOTE_A4,8, NOTE_B4,8, NOTE_CS5,8, NOTE_D5,8, NOTE_B4,8, //35
  REST,8, NOTE_E5,8, NOTE_FS5,8, NOTE_E5,-4, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_FS5,-8, NOTE_FS5,-8, NOTE_E5,-4, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_E5,-8, NOTE_E5,-8, NOTE_D5,-8, NOTE_CS5,16, NOTE_B4,8, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_D5,4, NOTE_E5,8, NOTE_CS5,-8, NOTE_B4,16, NOTE_A4,4, NOTE_A4,8, 

   NOTE_E5,4, NOTE_D5,2, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16, //40
  NOTE_FS5,-8, NOTE_FS5,-8, NOTE_E5,-4, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_A5,4, NOTE_CS5,8, NOTE_D5,-8, NOTE_CS5,16, NOTE_B4,8, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_D5,4, NOTE_E5,8, NOTE_CS5,-8, NOTE_B4,16, NOTE_A4,4, NOTE_A4,8,  
  NOTE_E5,4, NOTE_D5,2, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
   
  NOTE_FS5,-8, NOTE_FS5,-8, NOTE_E5,-4, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16, //45
  NOTE_A5,4, NOTE_CS5,8, NOTE_D5,-8, NOTE_CS5,16, NOTE_B4,8, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_D5,4, NOTE_E5,8, NOTE_CS5,-8, NOTE_B4,16, NOTE_A4,4, NOTE_A4,8,  
  NOTE_E5,4, NOTE_D5,2, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_FS5,-8, NOTE_FS5,-8, NOTE_E5,-4, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16, //45
  
  NOTE_A5,4, NOTE_CS5,8, NOTE_D5,-8, NOTE_CS5,16, NOTE_B4,8, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_D5,4, NOTE_E5,8, NOTE_CS5,-8, NOTE_B4,16, NOTE_A4,4, NOTE_A4,8, 

  NOTE_E5,4, NOTE_D5,2, REST,4
};

// sizeof gives the number of bytes, each int value is composed of two bytes (16 bits)
// there are two values per note (pitch and duration), so for each note there are four bytes
int notes = sizeof(melody) / sizeof(melody[0]) / 2;

// this calculates the duration of a whole note in ms
int tempo = 114;
int wholenote = (60000 * 4) / tempo;

int divider = 0, noteDuration = 0;

// Sync tracking variables
bool receivedAlarmTime = false;
bool receivedAlarmStatus = false;

// Blynk Time Input handler
BLYNK_WRITE(V0) {
  long startTimeInSecs = param[0].asLong();
  time_t alarmTime = startTimeInSecs;
  struct tm *tm = localtime(&alarmTime);
  
  // Subtract UTC offset from hour
  alarmHour = tm->tm_hour - UTC_OFFSET;
  if (alarmHour < 0) {
    alarmHour += 24;  // Handle wrap around for negative hours
  }
  alarmMinute = tm->tm_min;
  
  Serial.println("\n=== Alarm Time Sync/Update ===");
  Serial.print("Raw time received: ");
  Serial.print(tm->tm_hour);
  Serial.print(":");
  if (tm->tm_min < 10) Serial.print("0");
  Serial.println(tm->tm_min);
  
  Serial.print("Adjusted for UTC-");
  Serial.print(UTC_OFFSET);
  Serial.print(": ");
  Serial.print(alarmHour);
  Serial.print(":");
  if (alarmMinute < 10) Serial.print("0");
  Serial.println(alarmMinute);
  
  receivedAlarmTime = true;
  printSyncStatus();
}

// Alarm status handler
BLYNK_WRITE(V1) {
  alarmEnabled = param.asInt();
  Serial.println("\n=== Alarm Status Sync/Update ===");
  Serial.print("Alarm state received: ");
  Serial.println(alarmEnabled ? "ENABLED" : "DISABLED");
  
  // If alarm is enabled, store current time as sleep time
  if (alarmEnabled) {
    struct tm timeinfo;
    if (getLocalTime(&timeinfo)) {
      sleepTimeInfo.hour = timeinfo.tm_hour;
      sleepTimeInfo.minute = timeinfo.tm_min;
      sleepTimeInfo.day = timeinfo.tm_mday;
      sleepTimeInfo.month = timeinfo.tm_mon + 1;
      sleepTimeInfo.year = timeinfo.tm_year + 1900;
      sleepTimeSet = true;
      
      char currentTimeStr[20];
      snprintf(currentTimeStr, sizeof(currentTimeStr), "%02d:%02d %02d/%02d/%04d",
               sleepTimeInfo.hour,
               sleepTimeInfo.minute,
               sleepTimeInfo.day,
               sleepTimeInfo.month,
               sleepTimeInfo.year
      );
      Serial.print("Sleep time set to: ");
      Serial.println(currentTimeStr);
    }
  }
  
  receivedAlarmStatus = true;
  printSyncStatus();
}

void setup() {
  Serial.begin(115200);
  Serial.println("\nStarting Alarm System");
  
  // Initialize pins
  pinMode(JOYSTICK_BUTTON_PIN, INPUT_PULLUP);
  pinMode(SPEAKER_PIN, OUTPUT);
  
  // Connect to WiFi
  Serial.print("Connecting to WiFi");
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("\nWiFi Connected!");

  // Connect to Blynk
  Serial.print("Connecting to Blynk...");
  Blynk.begin(BLYNK_AUTH_TOKEN, ssid, "");
  Serial.println("Connected!");
  
  // Set up time sync
  configTime(UTC_OFFSET * 3600, 0, "pool.ntp.org", "time.nist.gov");
  
  // Request initial states from Blynk
  delay(1000);
  requestBlynkSync();
  
  // Print initial state
  Serial.println("\n=== Initial State ===");
  Serial.print("Alarm time: ");
  Serial.print(alarmHour);
  Serial.print(":");
  Serial.println(alarmMinute);
  Serial.print("Alarm enabled: ");
  Serial.println(alarmEnabled ? "YES" : "NO");
  Serial.println("===================");
}

void loop() {
  Blynk.run();
  
  // Check for alarm time match only if alarm is enabled
  if (alarmEnabled) {
    checkAlarm();
  }
  
  // Check if joystick is pressed and alarm is playing
  if (digitalRead(JOYSTICK_BUTTON_PIN) == HIGH && isAlarmPlaying) {
    turnOffAlarm();
  }
  
  delay(100);
}

void checkAlarm() {
  struct tm timeinfo;
  if (!getLocalTime(&timeinfo)) {
    Serial.println("Failed to obtain time");
    return;
  }
  
  if (timeinfo.tm_hour == alarmHour && 
      timeinfo.tm_min == alarmMinute && 
      !isAlarmPlaying) {
    playAlarm();
  }
}

void playAlarm() {
  Serial.println("Playing alarm");
  isAlarmPlaying = true;
  
  for (int thisNote = 0; thisNote < notes * 2; thisNote = thisNote + 2) {
    // Check if joystick button is pressed during playback
    if (digitalRead(JOYSTICK_BUTTON_PIN) == LOW) {
      turnOffAlarm();
      return;
    }
    
    // calculates the duration of each note
    divider = melody[thisNote + 1];
    if (divider > 0) {
      // regular note, just proceed
      noteDuration = (wholenote) / divider;
    } else if (divider < 0) {
      // dotted notes are represented with negative durations!!
      noteDuration = (wholenote) / abs(divider);
      noteDuration *= 1.5; // increases the duration in half for dotted notes
    }
    
    // we only play the note for 90% of the duration, leaving 10% as a pause
    tone(SPEAKER_PIN, melody[thisNote], noteDuration * 0.9);
    
    // Wait for the duration of the note plus a small pause
    delay(noteDuration);
    
    // Stop the tone
    noTone(SPEAKER_PIN);
  }
  
  // If alarm completed playing and wasn't turned off, restart it
  if (isAlarmPlaying) {
    playAlarm();
  }
}

void turnOffAlarm() {
  // Only proceed if alarm is actually playing
  if (!isAlarmPlaying) {
    return;
  }
  
  noTone(SPEAKER_PIN);
  isAlarmPlaying = false;
  
  // Turn off alarm in Blynk
  Blynk.virtualWrite(V1, 0);
  alarmEnabled = false;
  
  // Get current time
  struct tm timeinfo;
  if (getLocalTime(&timeinfo) && sleepTimeSet) {
    // Format current time
    char currentTimeStr[20];
    snprintf(currentTimeStr, sizeof(currentTimeStr), "%02d:%02d %02d/%02d/%04d",
             timeinfo.tm_hour,
             timeinfo.tm_min,
             timeinfo.tm_mday,
             timeinfo.tm_mon + 1,
             timeinfo.tm_year + 1900
    );
    
    Serial.print("Wake up time: ");
    Serial.println(currentTimeStr);
    
    // Calculate time difference
    float hoursSlept = 0;
    if (timeinfo.tm_mday == sleepTimeInfo.day && 
        (timeinfo.tm_mon + 1) == sleepTimeInfo.month && 
        timeinfo.tm_year + 1900 == sleepTimeInfo.year) {
      // Same day
      hoursSlept = (timeinfo.tm_hour - sleepTimeInfo.hour) + 
                   (timeinfo.tm_min - sleepTimeInfo.minute) / 60.0;
      
      // For Video
      hoursSlept =+ 8;
    } else {
      // Different day
      hoursSlept = (24 - sleepTimeInfo.hour) + timeinfo.tm_hour + 
                   (timeinfo.tm_min - sleepTimeInfo.minute) / 60.0;
    }
    
    if (hoursSlept < 0) {
      hoursSlept += 24;  // Adjust for crossing midnight
    }
    
    Serial.print("Hours slept: ");
    Serial.println(hoursSlept, 2);
    
    // Send notification to Make.com with hours slept
    String payload = String(hoursSlept, 2);
    sendMakeNotification(payload);
  }
  
  // Reset sleep time tracking
  sleepTimeSet = false;
}

void sendMakeNotification(String payload) {
  if (WiFi.status() == WL_CONNECTED) {
    HTTPClient http;
    String url = "Enter Webhook URL" + payload;
    http.begin(url);
    http.addHeader("Content-Type", "application/json");
    
    // Send POST request
    int httpResponseCode = http.POST("{}");
    
    if (httpResponseCode > 0) {
      Serial.print("Make.com notification sent. Response code: ");
      Serial.println(httpResponseCode);
    } else {
      Serial.print("Error sending notification. Error code: ");
      Serial.println(httpResponseCode);
    }
    
    http.end();
  }
}

void printSyncStatus() {
  if (receivedAlarmTime && receivedAlarmStatus) {
    Serial.println("\n=== Sync Complete ===");
    Serial.println("Final settings:");
    Serial.print("Alarm time: ");
    Serial.print(alarmHour);
    Serial.print(":");
    if (alarmMinute < 10) Serial.print("0");
    Serial.println(alarmMinute);
    Serial.print("Alarm status: ");
    Serial.println(alarmEnabled ? "ENABLED" : "DISABLED");
    Serial.println("==================");
  }
}

void requestBlynkSync() {
  Serial.println("\n=== Requesting initial sync from Blynk ===");
  receivedAlarmTime = false;
  receivedAlarmStatus = false;
  Blynk.syncVirtual(V0);  // Get alarm time
  Blynk.syncVirtual(V1);  // Get alarm status
} 